; -------------------------------------
; Z80 SRAM K1-bus micro computer board 
; with Interpreter Virtual Machine 
; (c) 2016 - 2017 kio@little-bat.de
; -------------------------------------

; Registers:
;
; AF =
; BC = VM PC
; DE = VM vstack TOP value
; HL = 
; IX = VM rstack
; SP = VM vstack
; IY = 
; AF' .. HL' = not used

;#assert VM_BC_IS_VPC	    == 1
;#assert VM_SP_IS_VSTACK     == 1
;#assert VM_IX_IS_RSTACK     == 1
;#assert VM_HAS_HSTACK	    == 0
;#assert VM_DE_IS_TOP	    == 1
;#assert VM_ALWAYS_EI	    == 1
;#assert VM_ALWAYS_DI	    == 0
;#assert HEAP_MALLOC_MFREE   == 0
;#assert HEAP_RETAIN_RELEASE == 1
#assert HEAP_USES_HANDLES   == 1


#local
#code HW_DRIVER


; –––––––––––––––––––––––––––––––––––––––
; system timer:
;
; note:	the system timer interrupt is also used to read and write to the sio channels.
; 	100 Hz is sufficient for 9600 baud: 100*10*16 = 16000 > 9600

XTAL = 7372000			; UART clock
timer_frequency = 100		; Hz

; note:	f = 50..1000Hz  <=>  period = 1000..20000µs  <=>  C/T_value = 230..4607
;
;	ctp_value   value.µs
;	--------- = -------	<=>  ctp_value * 32e6 = value.µs * XTAL <=> ctp_value = value.µs / (32e6/XTAL)
;	XTAL/16/2   1000000			

timer_value = 2304		; CTP value = XTAL / 32.0 / timer_frequency	
soll_period = 9215		; = 9215
real_period = 9216		; = 9215 * (timer_value*32) / (XTAL/timer_frequency)
				; -> every 9215 interrupts the systemtime is too low by 1


; –––––––––––––––––––––––––––––––––––––––
; system timer variables:
; the system timer variables are in _INITIALIZED, so that the #assert works:

#data DATA
systemtime:: 	ds  4		; uint32  time in millisec  (up to 49 days)
period::	ds  2 		; dw -soll_period	; uint16  Ausgleichzähler   
				; if left uninitialized, the first ~ 7 leaps are missed ~ 0.07sec
timestamp::	ds  4		; uint32  seconds since 1.1.2000
tsmsec::	ds  2		; dw -1000		; uint16  systemtime.lo der nächsten vollen Minute
				; if left uninitialized, the first second is incremented immediately

#assert hi(systemtime) == hi(period+1)


; –––––––––––––––––––––––––––––––––––––––
#if 0 ; while not defined in sio.cc
#data DATA
#include "sio_defines.s"
#code PAGE0
;imr_value::	db  IMR_value 	; current value of interrupt mask register IMR
;opcr_value::	db  OPCR_value	; current value of output port control register OPCR
#data DATA
;sio1A::		ds  2		; -> -> SioData
;sio1B::		ds  2		; -> -> SioData
;sio_handlers::	ds  2		; -> -> FD_Handlers

#code CONST
; Definition of struct "SerialDevice":
FD.handlers	equ	0	;	FDHandlers*
FD.nodetype	equ	2	;	uint8
FD.name		equ	3	;	char*
SerialDevice.hw_handshake  = 5	;	bool
SerialDevice.sw_handshake  = 6	;	uint8
SerialDevice.clk_handshake = 8	;	uint8		   
SerialDevice.device	= 9	;	uint8	// device select mask
SerialDevice.channel	= 10	;	uint8	// 0 = channel A; 1 = channel B
SerialDevice.baudrate	= 11	;	uint8	// baudrate / 2400
SerialDevice.sreg	= 12	;	uint8	// io address of status register SRA or SRB incl. k1_wr_data
SerialDevice.rxreg	= 13	;	uint8	// io address of receive register RXA or RXB incl. k1_wr_data
SerialDevice.txreg	= 14	;	uint8	// io address of transmit register TXA or TXB incl. k1_wr_data
;SerialDevice.ibusize	= 15	const
;SerialDevice.obusize	= 16	const
;SerialDevice.ibumask	= 17	const
;SerialDevice.obumask	= 18	const
SerialDevice.ibuwi	= 15	;	uint8	// input  buffer write index
SerialDevice.iburi	= 16	;	uint8	// input  buffer read index
SerialDevice.obuwi	= 17	;	uint8	// output buffer write index
SerialDevice.oburi	= 18	;	uint8	// output buffer read index
SerialDevice.ibu	= 19	;	uint8*	// input  buffer
SerialDevice.obu	= 21	;	uint8*	// output buffer
; related constants:
SerialDevice.obusz	= 64	;	output buffer size; must be 64
SerialDevice.ibusz	= 64	;	input buffer size;  must be 64
SerialDevice.obumask	= SerialDevice.obusz-1;
SerialDevice.ibumask	= SerialDevice.ibusz-1;
; bits in SerialDevice.sw_handshake:
SerialDevice.mSWHSK_ENABLED = 1;
SerialDevice.mXOFF_RECEIVED = 2;
SerialDevice.mXOFF_SENT     = 4;
SerialDevice.bSWHSK_ENABLED = 0;
SerialDevice.bXOFF_RECEIVED = 1;
SerialDevice.bXOFF_SENT     = 2;
#endif



; –––––––––––––––––––––––––––––––––––––––––
; 88C192 UART registers:
;
; NOTE: registers and most constants are defined in "sio.cc" and "sio.h"!
;
; note:	don't forget to add k1_rd_data and k1_wr_data to the address 
;	in all IN and OUT instructions!

; buffer size:
obusz		= SerialDevice.obusz;		output buffer size: must be 2^3 … 2^7 (not 8!!)
ibusz		= SerialDevice.ibusz;		input buffer size:  must be 2^3 … 2^7
obumask		= obusz-1;
ibumask		= ibusz-1;

; struct SerialDevice: 
; contains all data for one channel
; offsets of data members:
fd_handlers::	equ FD.handlers;		FDHandlers*
fd_nodetype::	equ FD.nodetype;		uint8
fd_dev_name	equ FD.name;			char*
hw_handshake	equ SerialDevice.hw_handshake;	bool
sw_handshake	equ SerialDevice.sw_handshake;	uint8
clk_handshake	equ SerialDevice.clk_handshake;	uint8		   
device		equ SerialDevice.device;	uint8	// device select mask
channel		equ SerialDevice.channel;	uint8	// 0 = channel A; 1 = channel B
baudrate	equ SerialDevice.baudrate;	uint8	// baudrate / 2400
sr_addr		equ SerialDevice.sreg
rx_addr		equ SerialDevice.rxreg
tx_addr		equ SerialDevice.txreg
ibuwi		equ SerialDevice.ibuwi;		uint8	// input  buffer write index
iburi		equ SerialDevice.iburi;		uint8	// input  buffer read index
obuwi		equ SerialDevice.obuwi;		uint8	// output buffer write index
oburi		equ SerialDevice.oburi;		uint8	// output buffer read index
ibu		equ SerialDevice.ibu;		uint8[]	// input  buffer
obu		equ SerialDevice.obu;		uint8[]	// output buffer
sizeof_SioData	equ obu + 2

; bits in SerialDevice.sw_handshake:
bSWHSK		equ SerialDevice.bSWHSK_ENABLED	
bXOFF_RCVD	equ SerialDevice.bXOFF_RECEIVED
bXOFF_SENT	equ SerialDevice.bXOFF_SENT


; ––––––––––––––––––––––––––––––––––––––––––
; init data for sio and system timer:
; used by boot code to init the 88C192 UART

#code CONST
sio1_init_data::

    ; select SIO Board:
	db	k1_wr_sel, 	 mSIO_SELECT

uart_init_data::

    ; config registers:
	db	k1_wr_data+OPCR, OPCR_value	; 0x00
	db	k1_wr_data+SOPR, SOPR_value	; 0xff
	db	k1_wr_data+ROPR, ROPR_value	; 0x00
	;db	k1_wr_data+ACR,  ACR_value

    ; timer interrupt:
	db	k1_wr_data+CTPU, hi(timer_value) ; set C/T preload value
	db	k1_wr_data+CTPL, lo(timer_value)
	db	k1_wr_data+ACR,  ACR_value	; set clock source and C/T mode

    ; channel A:
	; set baudrate:
	db	k1_wr_data+CSRA, 0x99         	; Baudrate 9600
	db	k1_wr_data+CRA,  0xB0		; point MRx to MR0
	db	k1_wr_data+MR0A, MR0_value + 0 	; WDT, BdRateTable, RX und TX buffer trigger level
	; set hw flow control      	
	db	k1_wr_data+MR1A, MR1_value     	; OP0|1: receiver RTS flow control		
	db	k1_wr_data+MR2A, MR2_value     	; IP0|1: transmitter CTS flow control
	; reset channel
	db	k1_wr_data+CRA,  0x20          	; reset receiver
	db	k1_wr_data+CRA,  0x30          	; reset transmitter 
	db	k1_wr_data+CRA,  0x40 		; reset error status: Clears channel, break, 
						;	parity, and over-run error bits in SR
	db	k1_wr_data+CRA,  1<<0 + 1<<2   	; enable receiver and transmitter

    ; channel B:
	; set baudrate
	db	k1_wr_data+CSRB, 0x99         	; Baudrate 9600
	db	k1_wr_data+CRB,  0xB0		; point MRx to MR0
	db	k1_wr_data+MR0B, MR0_value + 0 	; WDT, BdRateTable, RX und TX buffer trigger level
	; set hw flow control      	
	db	k1_wr_data+MR1B, MR1_value     	; OP0|1: receiver RTS flow control		
	db	k1_wr_data+MR2B, MR2_value     	; IP0|1: transmitter CTS flow control
	; reset channel
	db	k1_wr_data+CRB,  0x20          	; reset receiver
	db	k1_wr_data+CRB,  0x30          	; reset transmitter 
	db	k1_wr_data+CRB,  0x40 		; reset error status: Clears channel, break, 
						;	parity, and over-run error bits in SR
	db	k1_wr_data+CRB,  1<<0 + 1<<2   	; enable receiver and transmitter
	
    ; enable interrupts in the UART:
	db	k1_wr_data+IMR,  IMR_value	

    ; store test value:
	db	k1_wr_data+GPR,	 123

uart_init_data_count:: equ ($ - uart_init_data) / 2
sio1_init_data_count:: equ ($ - sio1_init_data) / 2

	
#if 0		; sio1A and sio1B are initialized in sio.cc
#code GSINIT
; ––––––––––––––––––––––––––––––––––––––––––––––––––––
; Note: the UART was initialized early in firmware.asm	
; we still need to allocated the SioData structs:
	
	; allocate SioData for channel A:
	ld	de,sizeof_SioData
	call	new_data_with_size		; zero cleared
	push	de
	call	new_handle_for_ptr
	ld	(sio1A),de			; handle
	
	; initialize data:
	pop	iy		       		; iy -> SioData
	ld	(iy+fd_nodetype), t_ser 	; t_ser = 8
	ld	(iy+device), mSIO_SELECT	; device select mask
	;ld	(iy+channel),0	       		; 0 = channel A; 1 = channel B
	ld	(iy+baudrate), 9600/2400	; baudrate / 2400

	ld	(iy+sr_addr), k1_rd_data + SRA
	ld	(iy+tx_addr), k1_wr_data + TXA
	ld	(iy+rx_addr), k1_rd_data + RXA

	ld	de, ibusz
	call	new_data_with_size
	call	new_handle_for_ptr
	ld	(iy+ibu+0),e			; uint8[ibusz]	
	ld	(iy+ibu+1),d

	ld	de, obusz
	call	new_data_with_size
	call	new_handle_for_ptr
	ld	(iy+obu+0),e			; uint8[obusz]	
	ld	(iy+obu+1),d

	;ld	(iy+fd_handlers+0),lo(sio_handlers)	TODO
	;ld	(iy+fd_handlers+1),hi(sio_handlers)
	; TODO: retain
	; TODO: name 
	
	; allocate SioData for channel B:
	ld	de,sizeof_SioData
	call	new_data_with_size	
	push	de
	call	new_handle_for_ptr
	ld	(sio1B),de			; handle
	                                	
	; initialize data:              	
	pop	iy				; iy -> SioData
	ld	(iy+fd_nodetype),t_ser		; t_ser = 8
	ld	(iy+device),mSIO_SELECT		; device select mask
	ld	(iy+channel),1			; 0 = channel A; 1 = channel B
	ld	(iy+baudrate),9600/2400		; baudrate / 2400

	ld	(iy+sr_addr), k1_rd_data + SRB
	ld	(iy+tx_addr), k1_wr_data + TXB
	ld	(iy+rx_addr), k1_rd_data + RXB

	ld	de, ibusz
	call	new_data_with_size
	call	new_handle_for_ptr      	
	ld	(iy+ibu+0),e			; uint8[ibusz]	input buffer
	ld	(iy+ibu+1),d

	ld	de, obusz
	call	new_data_with_size
	call	new_handle_for_ptr
	ld	(iy+obu+0),e			; uint8[obusz]	output buffer
	ld	(iy+obu+1),d

	;ld	(iy+fd_handlers+0),lo(sio_handlers)	TODO
	;ld	(iy+fd_handlers+1),hi(sio_handlers)
	; TODO: retain
	; TODO: name 

	; register enabled interrupt for the board:
	ld	hl, k1_irpt_mask
	res	bSIO_SELECT, (hl)
#endif


#code HW_DRIVER

; –––––––––––––––––––––––––––––––––––––––––––––
; Interrupt Handler:
; called with interrupts disabled 
; the device is not yet selected
;
; registers exxed
;
; do: 	increment system timer
;    	check UART channel A and B

sio_irpt::
	;ex	af,af'		; done by irpt handler
	;exx			; done by irpt handler
	push	ix

; select UART:
	ld	a,mSIO_SELECT
	out	(k1_wr_sel),a

; increment system timer:  
	; the system timer is the only enabled interrupt in the UART,
	; so we don't need to test whether it's interrupt is pending.
	call 	sio_timer
	
; check UART channel A:
	ld	hl,(sio1A)	; hl -> struct SioData handle
	call	sio_irpt_channel

; check UART channel B:
	ld	hl,(sio1B)	; hl -> struct SioData handle
	call	sio_irpt_channel

; exit:
	pop	ix
	exx
	ex	af,af'
	ei
	ret
  
 
; –––––––––––––––––––––––––––––––––––––––––––––
; handle interrupt for one channel:
; in:	hl = SioData¢ -> handle -> struct SioData data
;	note: register b is not used
;
;  1. Receive data
;  2. Send xon/xoff
;  3. Send data

sio_irpt_channel:
#assert HEAP_USES_HANDLES
	
	ld	a,(hl++)
	ld	ixl,a
	ld	a,(hl)
	ld	ixh,a		; ix -> SioData struct data
	
; Receive data:
	ld	l,(ix+ibu+0)
	ld	h,(ix+ibu+1)	; hl -> -> ibu data
	ld	e,(hl++)
	ld	d,(hl)		; de -> ibu data

; while in(SR) & mSR_RX_ready && ix.ibuwi - ix.iburi < ibusz
1$:	ld	c,(ix+sr_addr)
	in	c,(c) 		; c = in(SR)
	bit	bSR_RX_ready,c
	jr	z,10$		; rx fifo empty

; rx data available:
; check ibu[] free:
	ld	a,(ix+ibuwi)	
	sub	(ix+iburi)
	cp	ibusz	
	jr	nc,10$		; ibu[] full

; a = byte = in(RX)
	ld	c,(ix+rx_addr)
	in	a,(c) 		; a = byte = in(RX)

; in sw hsk mode test for xon and xoff:
	bit	bSWHSK,(ix+sw_handshake)
	jr	z,2$		; no sw hsk 
	cp	xoff
	jr	z,3$
	cp	xon
	jr	z,4$
	
; store normal byte:
; ix.ibu[ix.ibuwi++ & ibumask] = byte
2$:	ld	c,a		; c = byte
	ld	a,(ix+ibuwi)
	inc	(ix+ibuwi)
	and	ibumask		; a = sio1A.ibu[i]
	ld	l,a
	ld	h,0
	add	hl,de
	ld	(hl),c
	jr	1$

; xoff or xon received:
3$:	set	bXOFF_RCVD,(ix+sw_handshake)
	jr	1$
4$:	res	bXOFF_RCVD,(ix+sw_handshake)
	jr	1$

; Send xon/xoff:

; if(ix.sw_handshake && in(SR) & mSR_TX_ready)
10$:	bit	bSR_TX_ready,c		; c = in(SR) 	still valid from above
	ret	z			; tx fifo full

	ld	l,(ix+obu+0)
	ld	h,(ix+obu+1)		; hl -> -> obu
	ld	e,(hl++)
	ld	d,(hl)			; de -> obu data

	ld	h,(ix+sw_handshake)
	bit	bSWHSK,h
	jr	z,21$			; no sw hsk

; a = ibu.used = ix.ibuwi - ix.iburi
	ld	a,(ix+ibuwi)
	sub	(ix+iburi)

	cp	ibusz/2			; used - low water mark
	jr	nc,11$			; used ≥ low water mark
; ibu[] < low water mark:
; send xon if xoff sent:
	bit	bXOFF_SENT,h		; h = (ix+sw_handshake)
	jr	z,18$			; xoff not sent
; send xon:
	res	bXOFF_SENT,h		; h = (ix+sw_handshake)		remember xoff not sent
	ld	a,xon
	jr	19$			; send xon

; ibu[] ≥ low water mark:
11$:	cp	ibusz*7/8		; used - high water mark
	jr	c,18$			; used < high water mark  (and ≥ low water mark)
; ibu[] ≥ high water mark:
; send xoff if not yet sent:
	bit	bXOFF_SENT,h		; h = (ix+sw_handshake)
	jr	nz,18$			; already sent
; send xoff:
	set	bXOFF_SENT,h		; h = (ix+sw_handshake) 	remember xoff sent
	ld	a,xoff
	;jr	19$			; send xoff
19$:	ld	(ix+sw_handshake),h
	ld	c,(ix+tx_addr)
	out	(c),a
; test for xoff received before sending data:
18$:	bit	bXOFF_RCVD,h
	ret	nz			; xoff received -> exit
	;jr	20$

; Send data:
  
; while(in(SR) & mSR_TX_ready && ix.obuwi!=ix.oburi)
20$: 	ld	c,(ix+sr_addr)
	in	a,(c) 			; a = in(SR)
	and	mSR_TX_ready
	ret	z			; tx fifo full

21$:	ld	a,(ix+oburi)
	cp	(ix+obuwi)
	ret 	z			; obu[] empty

; out(TX,ix.obu[ix.oburi++ & obumask]);
	;ld	a,(ix+oburi)
	inc	(ix+oburi)
	and	obumask
	ld	l,a
	ld	h,0
	add	hl,de
	ld	a,(hl)			; a = byte
	ld	c,(ix+tx_addr)
	out	(c),a
	jr	20$


; ––––––––––––––––––––––––––––––––––––––––––––––
sio_timer:: 

; switch off interrupt
; note: in timer mode the stop C/T command only resets interrupt and does not stop the C/T
	in	a,(k1_rd_data + SPCR)	


; –––––––––––––––––––––––––––––
; increment timestamp [seconds]

; if(systemtime + tsmsec >= 0)   
; <=> Δ.lo < 11 && Δ.hi == 0
; we test Δ.lo first, because it is more likely:

	ld	a,(tsmsec)	; tsmsec.lo
	ld	hl,systemtime	; systemtime.llo
	add	a,(hl)		;
	cp	11		; note: systemtime might be incremented by 11
	jr	c,8$		; Δ.lo < 11


; –––––––––––––––––––––––––––––
; increment systemtime [ms]

; systemtime += 10		; 1000 / timer_frequency

	;ld	hl,systemtime	; systemtime.llo
5$:	ld	a,10
	add	(hl)
	ld	(hl),a
	jr	nc,1$
2$:	inc	l
	inc	(hl)
	jr	z,2$

; period += 1;
; if(period < 65536) return

1$:	ld	l,lo(period)
	inc	(hl)
	ret	nz
	inc	l
	inc	(hl)
	ret	nz

; period -= soll_period;

	ld	(hl),hi(-soll_period)
	dec	l
	ld	(hl),lo(-soll_period)

; systemtime++;			

	ld	l,lo(systemtime)
4$:	inc	(hl)
	ret	nz
	inc	l
	jr	4$


; –––––––––––––––––––––––––––––
; increment timestamp (ctd.)

; test for Δ.hi == 0
8$:
;	ld	hl,(tsmsec)
;	ld	de,(systemtime)
;	add	hl,de
;	ld	a,h
;	and	a
;	ld	hl, systemtime	; hl -> systime.llo
;	jr	nz,5$

	ld	a,(tsmsec)	; redo calculation:
	add	a,(hl)		; we need the CY
	inc	hl
	ld	a,(tsmsec+1)
	adc	(hl)	
	dec	hl
	jr	nz,5$
	
; calc. time of next full second:
	ld	hl, (tsmsec)
	ld	de, -1000
	add	hl, de
	ld	(tsmsec), hl
; increment timestamp:
	ld	hl, timestamp
6$:	inc	(hl)
	inc	hl
	jr	z,6$
	ld	hl, systemtime	; hl -> systime.llo
	jr	5$



;; –––––––––––––––––––––––––––––
;; uint8 sio_avail_in(SioData¢ channel)  
;; { 
;;	return channel.ibuwi - channel.iburi; 
;; }
;
;sio_avail_in__12SerialDeviceC_6uint16::
;sio_avail_in::
;	ex	hl,de		; hl -> -> channel
;	ld	e,(hl++)
;	ld	d,(hl)		; de -> channel
;	
;	ld	hl,ibuwi
;	add	hl,de		; hl -> channel.ibuwi
;	ld	a,(hl++)	; a=ibuwi, hl->iburi
;	sub	a,(hl)		; a=ibuwi-iburi
;	ld	e,a
;	jp	extend_pos

;; –––––––––––––––––––––––––––––
;; uint8 sio_avail_out(SioData¢ channel) 
;; { 
;;	return obusz - (channel.obuwi - channel.oburi); 
;; }
;
;sio_avail_out__12SerialDeviceC_6uint16::
;sio_avail_out::				
;	ex	hl,de		; hl -> -> channel
;	ld	e,(hl++)
;	ld	d,(hl)		; de -> channel
;
;	ld	a,obusz		; a=obusz
;	ld	hl,obuwi
;	add	hl,de		; hl -> channel.obuwi
;	sub	a,(hl++)	; a=obusz-obuwi, hl->oburi
;	add	a,(hl)		; a=obusz-obuwi+iburi
;	ld	e,a
;	jp	extend_pos
	

; –––––––––––––––––––––––––––––
; write 1 character
; blocking
;
; void sio_putc( SioData¢ channel, uint8 c )
; {
; 	// wait until byte available: 
; 	while(!sio_avail_out(channel)){ wait(); }
; 
; 	// store byte. increment obuwi _after_ writing the byte:
; 	channel.obu[channel.obuwi & obumask] = c;
; 	channel.obuwi++;
; }

;sio_putc::			
;	p_enter			; stack: channel¢, char
;	
;; while(!sio_avail_out(channel)){ wait(); }
;	jrp 2$
;1$:	dw HALT
;2$:	over			; stack: channel¢, char, channel¢
;	dw sio_avail_out	; stack: channel¢, char, n
;	jrm0 1$			; stack: channel¢, char
;
;; channel.obu[channel.obuwi & obumask] = c;
;	over			; stack: channel¢, char, channel¢
;	iget obu		; stack: channel¢, char, channel.obu	Note: pointer no array!!
;	lget 4			; stack: channel¢, char, channel.obu, channel¢
;	igetu8 obuwi		; stack: channel¢, char, channel.obu, obuwi
;	ivalu8 obumask		; stack: channel¢, char, channel.obu, obuwi, obumask
;	dw AND			; stack: channel¢, char, channel.obu, obuwi&obumask
;	dw ADD			; stack: channel¢, char, &channel.obu[obuwi&obumask]
;	poke8			; stack: channel¢
;
;; channel.obuwi++;
;	ivar obuwi		; stack: &channel.obuwi
;	dw PPSET8		; stack: -
;	return


; –––––––––––––––––––––––––––––
; get 1 character
; blocking
; 
; uchar sio_getc( SioData¢ channel )
; {
; 	// wait until byte available: 
; 	while(!sio_avail_in(channel)){ wait(); }
; 
; 	// get byte. increment iburi _after_ reading the byte:
; 	uchar c = channel.ibu[channel.iburi & ibumask];
; 	channel.iburi++;
; 	return c;
; }

;sio_getc::
;	p_enter
;	
;; while(!sio_avail_in(channel)){ wait(); }
;	jrp 2$
;1$:	dw HALT
;2$:	dup			; stack: channel¢, channel¢
;	dw sio_avail_in		; stack: channel¢, n
;	jrm0 1$			; stack: channel¢
;
;; uchar c = channel.ibu[channel.iburi & ibumask];
;	dup			; stack: channel¢, channel¢
;	iget ibu		; stack: channel¢, channel.ibu		Note: pointer no array!
;	over			; stack: channel¢, channel.ibu, channel¢
;	igetu8 iburi		; stack: channel¢, channel.ibu, iburi
;	ivalu8 ibumask		; stack: channel¢, channel.ibu, iburi ibumask
;	dw AND			; stack: channel¢, channel.ibu, iburi&ibumask
;	dw ADD			; stack: channel¢, &channel.ibu[iburi&ibumask]
;	dw PEEKu8		; stack: channel¢, char
;
;; channel.iburi++;
;	swap			; stack: char, channel¢
;	ivar iburi		; stack: char, &channel.iburi
;	dw PPSET8		; stack: char
;	return
	

#endlocal











